unit UFMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, StdCtrls, ExtCtrls, Buttons, ImgList, Menus, Math;

CONST
  MESSAGE_BOX_POS = 200;

type
  TFieldStatus = (fsEmpty, fsX, fsO, fsTemporary, fsWin);
  TPlaygroundArray = array [1..3,1..3] of TFieldStatus;
  TPlayerSign = (psX, psO);
  TWinningSet = (wsNone, wsHorizontal, wsVertical, wsLeftRight, wsRightLeft);
  TfrmMain = class(TForm)
    Panel1: TPanel;
    pStatus: TPanel;
    Help1: TMenuItem;
    About1: TMenuItem;
    Panel2: TPanel;
    Image18: TImage;
    imgHuman: TImage;
    sbSign: TSpeedButton;
    imgComputer: TImage;
    Image15: TImage;
    cbLevel: TComboBox;
    GroupBox1: TGroupBox;
    lblWinsHuman: TLabel;
    lblWinsComputer: TLabel;
    Label5: TLabel;
    Label8: TLabel;
    Bevel1: TBevel;
    lblGamesPlayed: TLabel;
    lblDraws: TLabel;
    Image3: TImage;
    Image4: TImage;
    Image5: TImage;
    Image6: TImage;
    sbReset: TSpeedButton;
    sbPlay: TSpeedButton;
    Label9: TLabel;
    Panel3: TPanel;
    dgPlayground: TDrawGrid;
    procedure cbLevelChange(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure About1Click(Sender: TObject);
    procedure sbResetClick(Sender: TObject);
    procedure sbPlayClick(Sender: TObject);
  published
    ImageList: TImageList;
    imgTick: TImage;
    ImageListSmall: TImageList;
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    Exit1: TMenuItem;
    procedure dgPlaygroundDrawCell(Sender: TObject; ACol, ARow: Integer;
      Rect: TRect; State: TGridDrawState);
    procedure FormCreate(Sender: TObject);
    procedure sbSignClick(Sender: TObject);
    procedure dgPlaygroundSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure Exit1Click(Sender: TObject);
  private
    FiGameCount: integer;
    FiWinsHuman: integer;
    FiWinsComputer: integer;
    FComputerPlays: TPlayerSign;
    FiDraws: integer;
    FbHumanOpponent: boolean;
    FWhoseTurnNow: TPlayerSign;
    iMoveCount: integer;
    FbGameOver: boolean;
    PlaygroundArray: TPlaygroundArray;
    WinnerArray: array[1..3, 1..3] of integer;
    Level: integer;
    procedure SetiGameCount(const Value: integer);
    property iGameCount: integer read FiGameCount write SetiGameCount;
    procedure SetiWinsComputer(const Value: integer);
    procedure SetiWinsHuman(const Value: integer);
    procedure SetComputerPlays(const Value: TPlayerSign);
    property ComputerPlays: TPlayerSign read FComputerPlays write SetComputerPlays;
    property iWinsHuman: integer read FiWinsHuman write SetiWinsHuman;
    property iWinsComputer: integer read FiWinsComputer write SetiWinsComputer;
    procedure SetiDraws(const Value: integer);
    property iDraws: integer read FiDraws write SetiDraws;
    procedure SetbHumanOpponent(const Value: boolean);
    procedure BestMoveCalculate(var aCol, aRow: integer);
    procedure ComputerMove;
    property bHumanOpponent: boolean read FbHumanOpponent write SetbHumanOpponent; //TODO: Human Vs. Human
    procedure SetWhoseTurnNow(const Value: TPlayerSign);
    procedure SetbGameOver(const Value: boolean);
    procedure ResetPlayground;
    property bGameOver: boolean read FbGameOver write SetbGameOver;
    property WhoseTurnNow: TPlayerSign read FWhoseTurnNow write SetWhoseTurnNow;
    procedure PlaygroundClicked(aCol, aRow: integer);
    function IsWinner(aCol, aRow: integer; aFieldStatus: TFieldStatus): TWinningSet;
    procedure WeHaveAWinner(aCol, aRow: integer; aWinningSet: TWinningSet);
    procedure NextPlayerTurn;
    procedure NewGame;
    procedure WeHaveADraw;
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

uses UFNewGame, UFAbout;

{$R *.DFM}

procedure TfrmMain.SetbGameOver(const Value: boolean);
begin
  FbGameOver := Value;
  sbPlay.Enabled := Value;
  sbSign.Enabled := Value;
  cbLevel.Enabled := Value;
end;

procedure TfrmMain.PlaygroundClicked(aCol, aRow: integer);
var
  _FieldStatus: TFieldStatus;
  _WinningSet: TWinningSet;
begin
  if bGameOver then Exit;
  if PlaygroundArray[aCol, aRow] = fsEmpty then
  begin
    inc(iMoveCount);
    if WhoseTurnNow = psX then _FieldStatus := fsX
    else _FieldStatus := fsO;
    PlaygroundArray[aCol, aRow] := _FieldStatus;
    dgPlayground.Refresh;
    _WinningSet := IsWinner(aCol, ARow, _FieldStatus);
    if _WinningSet <> wsNone then
    begin
      bGameOver := true;
      WeHaveAWinner(aCol, aRow, _WinningSet);
      if WhoseTurnNow = ComputerPlays then iWinsComputer := iWinsComputer + 1
      else iWinsHuman := iWinsHuman + 1;
    end
    else
      if iMoveCount = 9 then
      begin
        bGameOver := true;
        WeHaveADraw;
        iDraws := iDraws + 1;
      end
      else NextPlayerTurn;
  end;
end;

procedure TfrmMain.dgPlaygroundDrawCell(Sender: TObject; ACol,
  ARow: Integer; Rect: TRect; State: TGridDrawState);
var
  _Bitmap: TBitmap;
begin
  with dgPlayground.Canvas do
  begin
    Pen.Color := $0053A9FF;
    Pen.Width := 3;
    PolyLine([Point(Rect.Left-2, Rect.Top-1), Point(Rect.Right+1, Rect.Top-1),
      Point(Rect.Right+1, Rect.Bottom+1), Point(Rect.Left-1, Rect.Bottom+1),
      Point(Rect.Left-1, Rect.Top-2)]);
    Pen.Width := 1;
    Pen.Color := $00FF00FF;
    PolyLine([Point(Rect.Left-1, Rect.Top), Point(Rect.Right, Rect.Top),
      Point(Rect.Right, Rect.Bottom), Point(Rect.Left, Rect.Bottom),
      Point(Rect.Left, Rect.Top-1)]);
  end;
  case PlaygroundArray[ACol, ARow] of
    fsX:
      with dgPlayground.Canvas do
      begin
        _Bitmap := TBitmap.Create;
        ImageList.GetBitmap(1, _Bitmap);
        Draw(Rect.Left+3, Rect.Top+3, _Bitmap);
        _Bitmap.Free;
      end;
    fsO:
      with dgPlayground.Canvas do
      begin
        _Bitmap := TBitmap.Create;
        ImageList.GetBitmap(0, _Bitmap);
        Draw(Rect.Left+3, Rect.Top+3, _Bitmap);
        _Bitmap.Free;
      end;
  end;
  if WinnerArray[ACol, ARow] = 1 then
    with dgPlayground.Canvas do
    begin
      CopyMode := cmSrcAnd;
      Draw(Rect.Left, Rect.Top, imgTick.Picture.Bitmap);
    end;
end;

procedure TfrmMain.ResetPlayground;
var
  i, k: integer;
begin
  for i := 1 to 3 do
    for k := 1 to 3 do
    begin
      PlaygroundArray[i, k] := fsEmpty;
      WinnerArray[i, k] := 0;
    end;
  dgPlayground.Refresh;
  pStatus.Caption := '';
  pStatus.ParentColor := true;
end;

procedure TfrmMain.SetWhoseTurnNow(const Value: TPlayerSign);
begin
  FWhoseTurnNow := Value;
  if Value = ComputerPlays then pStatus.Caption := 'Computer`s turn'
  else pStatus.Caption := 'Your turn';
  Application.ProcessMessages;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  dgPlayground.ColWidths[0] := 0;
  dgPlayground.RowHeights[0] := 0;
  bHumanOpponent := false;
  iGameCount := 0;
  iWinsHuman := 0;
  iWinsComputer := 0;
  iDraws := 0;
  bGameOver := true;
  ComputerPlays := psO;
  Randomize;
  Level := cbLevel.ItemIndex;
  dgPlayground.DoubleBuffered := true;
end;

function TfrmMain.IsWinner(aCol, aRow: integer; aFieldStatus: TFieldStatus): TWinningSet;
var
  i, _Count: integer;
begin
  _Count := 1;
  Result := wsNone;
  for i := aCol + 1 to aCol + 2 do
    if PlaygroundArray[(i - 1) mod 3 + 1, aRow] in [aFieldStatus, fsTemporary] then inc(_Count);
  if _Count = 3 then
  begin
    Result := wsHorizontal;
    Exit;
  end;
  _Count := 1;
  for i := aRow + 1 to aRow + 2 do
    if PlaygroundArray[aCol, (i - 1) mod 3 + 1] in [aFieldStatus, fsTemporary] then inc(_Count);
  if _Count = 3 then
  begin
    Result := wsVertical;
    Exit;
  end;
  _Count := 1;
  if aRow = aCol then
  begin
    for i := aRow + 1 to aRow + 2 do
      if PlaygroundArray[(i-1) mod 3 + 1, (i-1) mod 3 + 1] in [aFieldStatus, fsTemporary]
      then inc(_Count);
  end;
  if _Count = 3 then
  begin
    Result := wsLeftRight;
    Exit;
  end;
  _Count := 1;
  if aRow + aCol = 4 then
  begin
    for i := aRow + 1 to aRow + 2 do
      if PlaygroundArray[3 - (i - 1) mod 3, (i-1) mod 3 + 1] in [aFieldStatus, fsTemporary]
      then inc(_Count);
  end;
  if _Count = 3 then
  begin
    Result := wsRightLeft;
    Exit;
  end;
end;

procedure TfrmMain.NextPlayerTurn;
begin
  if WhoseTurnNow = psO then WhoseTurnNow := psX
  else WhoseTurnNow := psO;
  if not bHumanOpponent and (WhoseTurnNow = ComputerPlays) then ComputerMove;
end;

procedure TfrmMain.WeHaveAWinner(aCol, aRow: integer; aWinningSet: TWinningSet);
var
  i: integer;
begin
  case aWinningSet of
    wsHorizontal:
      begin
        for i := aCol to aCol + 2 do
          WinnerArray[1 + (i - 1) mod 3, aRow] := 1;
      end;
    wsVertical:
      with dgPlayground do
      begin
        for i := aRow to aRow + 2 do
          WinnerArray[aCol, 1 + (i - 1) mod 3] := 1;
      end;
    wsLeftRight:
      with dgPlayground do
      begin
        for i := aRow to aRow + 2 do
          WinnerArray[1 + (i - 1) mod 3, 1 + (i - 1) mod 3] := 1;
      end;
    wsRightLeft:
      with dgPlayground do
      begin
        for i := aRow to aRow + 2 do
        begin
          WinnerArray[1 + (i - 1) mod 3, 3 - (i - 1) mod 3] := 1;
        end;
      end;
  end;
  dgPlayground.Refresh;
  if WhoseTurnNow = ComputerPlays then
  begin
    pStatus.Color := $00FF6464;
    pStatus.Caption := 'You lose';
  end
  else
  begin
    pStatus.Color := $000080FF;
    pStatus.Caption := 'You win!';
  end;
end;

procedure TfrmMain.NewGame;
begin
  bGameOver := false;
  ResetPlayground;
  iGameCount := iGameCount + 1;
  iMoveCount := 0;
  pStatus.ParentColor := true;
  if iGameCount mod 2 = 0 then WhoseTurnNow := ComputerPlays
  else
    if ComputerPlays = psX then WhoseTurnNow := psO
    else WhoseTurnNow := psX;

  if not bHumanOpponent and (WhoseTurnNow = ComputerPlays) then ComputerMove;
end;

procedure TfrmMain.WeHaveADraw;
begin
  pStatus.Color := $0000FBFB;
  pStatus.Caption := 'It`s a Draw';
end;

procedure TfrmMain.SetbHumanOpponent(const Value: boolean);
begin
  FbHumanOpponent := Value;
end;

procedure TfrmMain.ComputerMove;
var
  _iCol, _iRow: integer;
begin
  BestMoveCalculate(_iCol, _iRow);
  Sleep(500); //Computer thinking time
  PlayGroundClicked(_iCol, _iRow);
end;

procedure TfrmMain.BestMoveCalculate(var aCol, aRow: integer);
var
  i, k, m, n, _WinPossibilities: integer;
  _FieldStatus: TFieldStatus;
begin
  if WhoseTurnNow = psO then _FieldStatus := fsO
  else _FieldStatus := fsX;
  //Can we win now?
  //On the novice level you have a 5% chance computer misses its winning move
  if (Level > 0) or (Random >= 0.05) then
  for i := 1 to 3 do
    for k := 1 to 3 do
      if (PlaygroundArray[i, k] = fsEmpty) and (IsWinner(i, k, _FieldStatus) <> wsNone) then
      begin
        aCol := i;
        aRow := k;
        Exit;
      end;
  if WhoseTurnNow = psO then _FieldStatus := fsX
  else _FieldStatus := fsO;
  //Can we block opponent's winning move?
  //On the novice level you have a 10% chance computer misses opponent's winning move
  if (Level > 0) or (Random >= 0.1) then
  for i := 1 to 3 do
    for k := 1 to 3 do
      if (PlaygroundArray[i, k] = fsEmpty) and (IsWinner(i, k, _FieldStatus) <> wsNone) then
      begin
        aCol := i;
        aRow := k;
        Exit;
      end;
  if WhoseTurnNow = psO then _FieldStatus := fsO
  else _FieldStatus := fsX;
  //Can we create a double chance situation? Advanced and expert level.
  if Level >= 1 then
    for m := 1 to 3 do
      for n := 1 to 3 do
      begin
        if PlaygroundArray[m, n] = fsEmpty then
        begin
          PlaygroundArray[m, n] := fsTemporary;
          aCol := m;
          aRow := n;
          _WinPossibilities := 0;
          for i := 1 to 3 do
            for k := 1 to 3 do
              if (PlaygroundArray[i, k] = fsEmpty) and (IsWinner(i, k, _FieldStatus) <> wsNone) then
                Inc(_WinPossibilities);
          PlaygroundArray[m, n] := fsEmpty;
          if _WinPossibilities > 1 then Exit;
        end;
      end;
  if WhoseTurnNow = psO then _FieldStatus := fsX
  else _FieldStatus := fsO;
  //Can we block opponent's double chance situation? Expert level only.
  if Level >= 2 then
    for m := 1 to 3 do
      for n := 1 to 3 do
      begin
        if PlaygroundArray[m, n] = fsEmpty then
        begin
          PlaygroundArray[m, n] := fsTemporary;
          aCol := m;
          aRow := n;
          _WinPossibilities := 0;
          for i := 1 to 3 do
            for k := 1 to 3 do
              if (PlaygroundArray[i, k] = fsEmpty) and (IsWinner(i, k, _FieldStatus) <> wsNone) then
                Inc(_WinPossibilities);
          PlaygroundArray[m, n] := fsEmpty;
          if _WinPossibilities > 1 then Exit;
        end;
      end;
  //Let's just make a random move
  i := 1 + Random(3);
  k := 1 + Random(3);
  while PlaygroundArray[i, k] <> fsEmpty do
  begin
    n := 1;
    i := 1 + i mod 3;
    while (PlaygroundArray[i, k] <> fsEmpty) and (n <= 3) do
    begin
      k := 1 + k mod 3;
      Inc(n);
    end;
  end;
  aCol := i;
  aRow := k;
end;

procedure TfrmMain.SetiDraws(const Value: integer);
begin
  FiDraws := Value;
  lblDraws.Caption := IntToStr(Value);
end;

procedure TfrmMain.sbSignClick(Sender: TObject);
begin
  if ComputerPlays = psX then ComputerPlays := psO
  else ComputerPlays := psX;
end;

procedure TfrmMain.SetiWinsComputer(const Value: integer);
begin
  FiWinsComputer := Value;
  lblWinsComputer.Caption := IntToStr(Value);
end;

procedure TfrmMain.SetiWinsHuman(const Value: integer);
begin
  FiWinsHuman := Value;
  lblWinsHuman.Caption := IntToStr(Value);
end;

procedure TfrmMain.dgPlaygroundSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  CanSelect := false;
  if not bGameOver and not (WhoseTurnNow = ComputerPlays) then PlayGroundClicked(ACol, ARow);
end;

procedure TfrmMain.SetComputerPlays(const Value: TPlayerSign);
begin
  FComputerPlays := Value;
  if FComputerPlays = psX then
  begin
    ImageListSmall.GetIcon(1, imgHuman.Picture.Icon);
    ImageListSmall.GetIcon(0, imgComputer.Picture.Icon);
  end
  else
  begin
    ImageListSmall.GetIcon(0, imgHuman.Picture.Icon);
    ImageListSmall.GetIcon(1, imgComputer.Picture.Icon);
  end;
end;

procedure TfrmMain.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TfrmMain.SetiGameCount(const Value: integer);
begin
  FiGameCount := Value;
  lblGamesPlayed.Caption := IntToStr(Value);
  sbReset.Enabled := Value > 0;
end;

procedure TfrmMain.cbLevelChange(Sender: TObject);
begin
  Level := cbLevel.ItemIndex;
end;

procedure TfrmMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  CanClose := bGameOver or (MessageDlgPos('Quit current game?', mtConfirmation, mbOKCancel, 0,
    Left + MESSAGE_BOX_POS, Top + MESSAGE_BOX_POS) = mrOk);
end;

procedure TfrmMain.About1Click(Sender: TObject);
begin
  with TAboutBox.Create(Self) do
  begin
    ShowModal;
    Free;
  end;
end;

procedure TfrmMain.sbResetClick(Sender: TObject);
begin
  if (MessageDlgPos('Clear all scores?', mtConfirmation, mbOKCancel, 0,
    Left + MESSAGE_BOX_POS, Top + MESSAGE_BOX_POS) = mrOk) then
  begin
    iGameCount := 0;
    iWinsHuman := 0;
    iWinsComputer := 0;
    iDraws := 0;
    bGameOver := true;
    ResetPlayground;
  end;
end;

procedure TfrmMain.sbPlayClick(Sender: TObject);
begin
  NewGame;
end;

end.
